/**
 * Copyright IBM Corporation 2016
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/

import Foundation

extension NSDate {

    /**
     Method returns a NSCalendar Unit

     - returns: NSCalendarUnit
     */
    private class func componentFlags() -> NSCalendarUnit {
        return NSCalendarUnit.Year.union(.Month).union(.Day).union(.WeekOfYear).union(.Hour).union(.Minute).union(.Second).union(.Weekday).union(.WeekdayOrdinal)
    }

    /**
     Method returns the components of the NSCalendar

     - parameter fromDate: NSDate

     - returns: NSDateComponents
     */
    private class func components(fromDate fromDate: NSDate) -> NSDateComponents {
        return NSCalendar.currentCalendar().components(NSDate.componentFlags(), fromDate: fromDate)
    }

    /**
     Method returns the NSDate components

     - returns: NSDateComponents
     */
    private func components() -> NSDateComponents {
        return NSDate.components(fromDate: self)
    }

    /**
     Method takes in a time interval since reference date and converts it to the number of seconds, minutes, hours, or weeks since the photo was taken

     - parameter timeInterval: NSTimeInterval

     - returns: String
     */
    class func timeStringSinceIntervalSinceReferenceDate(timeInterval: NSTimeInterval) -> String {

        let postedDate = NSDate(timeIntervalSinceReferenceDate: timeInterval)

        return timeSinceDateString(postedDate)

    }

    /**
     Method returns the number of years since the date parameter

     - parameter date: NSDate

     - returns: Int
     */
    private class func yearsFrom(date: NSDate) -> Int {
        return NSCalendar.currentCalendar().components(.Year, fromDate: date, toDate: NSDate(), options: []).year
    }

    /**
     Method returns the number of months since the date parameter

     - parameter date: NSDate

     - returns: Int
     */
    private class func monthsFrom(date: NSDate) -> Int {
        return NSCalendar.currentCalendar().components(.Month, fromDate: date, toDate: NSDate(), options: []).month
    }

    /**
     Method returns the number of weeks since the date parameter

     - parameter date: NSDate

     - returns: Int
     */
    private class func weeksFrom(date: NSDate) -> Int {
        return NSCalendar.currentCalendar().components(.WeekOfYear, fromDate: date, toDate: NSDate(), options: []).weekOfYear
    }

    /**
     Method returns the number of days since the date parameter

     - parameter date: NSDate

     - returns: Int
     */
    private class func daysFrom(date: NSDate) -> Int {
        return NSCalendar.currentCalendar().components(.Day, fromDate: date, toDate: NSDate(), options: []).day
    }

    /**
     Method returns the number of hours since the date parameter

     - parameter date: NSDate

     - returns: Int
     */
    private class func hoursFrom(date: NSDate) -> Int {
        return NSCalendar.currentCalendar().components(.Hour, fromDate: date, toDate: NSDate(), options: []).hour
    }

    /**
     Method returns the number of minutes since the date parameter

     - parameter date: NSDate

     - returns: Int
     */
    private class func minutesFrom(date: NSDate) -> Int {
        return NSCalendar.currentCalendar().components(.Minute, fromDate: date, toDate: NSDate(), options: []).minute
    }

    /**
     Method returns the number of seconds since the date parameter

     - parameter date: NSDate

     - returns: Int
     */
    private class func secondsFrom(date: NSDate) -> Int {
        return NSCalendar.currentCalendar().components(.Second, fromDate: date, toDate: NSDate(), options: []).second
    }

    /**
     Method returns the time since the data parameter as a string

     - parameter date: NSDate

     - returns: String
     */
    class func timeSinceDateString(date: NSDate) -> String {
        if weeksFrom(date)   > 0 { return "\(weeksFrom(date))" + NSLocalizedString("w", comment: "first letter of the word week")}
        if daysFrom(date)    > 0 { return "\(daysFrom(date))" +  NSLocalizedString("d", comment: "first letter of the word day")}
        if hoursFrom(date)   > 0 { return "\(hoursFrom(date))" +  NSLocalizedString("h", comment: "first letter of the word hour")}
        if minutesFrom(date) > 0 { return "\(minutesFrom(date))" +  NSLocalizedString("m", comment: "first letter of the word minutes")}
        if secondsFrom(date) > 0 { return "\(secondsFrom(date))" +  NSLocalizedString("s", comment: "first letter of the word seconds")}
        return NSLocalizedString("now", comment: "word representing this moment in time")
    }
}
